;; =============================================================================
;; MODULE: manual.el -- Manual alignment environment
;;
;; PROGRAMMER: Michel Simard -- TAO
;;
;; COMMENTS: 
;;
;;    Copyright (C) 1995, Center for Information Technology Innovation
;;    (CITI) -- All rights reserved.
;;
;; =============================================================================




;; =============================================================================
;; Alignment structure
;;
;; ALIGNMENT : (BUFFER(1) BUFFER(2)
;;		(COUPLE(1) COUPLE(2)...) 
;;		SAVE-TO-FILE
;;		MODIFIED-FLAG
;;		BACKED-UP-FLAG)
;; COUPLE    : (START(1) END(1) START(2) END(2) COMMENT)
;;
;; OK, BUFFER(X) are the buffers containing the aligned texts.  Each COUPLE
;; represents a pair of aligned regions in these texts.  An alignment will
;; always contain at least one couple, which then covers the two buffers
;; completely.  Until further notice, couples do not overlap or criss-cross,
;; and the alignment always covers the entirety of the two buffers.
;; Furthermore, the list of couples is sorted in increasing order of
;; START(1).  Start-positions are inclusive, end-positions are not. Start and
;; end positions are given as offsets from the beginning of the file (so that
;; the first character of the file is in position "0") -- hence, these must
;; be converted to and from Emacs "point" positions, by adding or
;; substracting 1.  SAVE-TO-FILE can be "nil", in the case of a new alignment.
;; =============================================================================

(defvar manual-alignment nil
  "The current alignment")
(defvar manual-couple nil
  "The index of the current couple of the current alignment")

(defun manual-get-buffer (half)
  ""
  (if manual-alignment (nth (1- half) manual-alignment)))

(defun manual-get-couple-list ()
  ""
  (if manual-alignment (nth 2 manual-alignment)))

(defun manual-get-save-to-file ()
  ""
  (if manual-alignment (nth 3 manual-alignment)))

(defun manual-get-modified-flag ()
  ""
  (if manual-alignment (nth 4 manual-alignment)))

(defun manual-get-backed-up-flag ()
  ""
  (if manual-alignment (nth 5 manual-alignment)))


(defun manual-get-couple-start (couple-no half)
  ""
  (let (couple)
    (if (and manual-alignment 
	     (<= 0 couple-no)
	     (setq couple (nth couple-no (manual-get-couple-list))))
	(1+ (nth (* 2 (1- half)) couple)))))

(defun manual-get-couple-end (couple-no half)
  ""
  (let (couple)
    (if (and manual-alignment 
	     (<= 0 couple-no)
	     (setq couple (nth couple-no (manual-get-couple-list))))
	(1+ (nth (1+ (* 2 (1- half))) couple)))))


;; -----------------------------------------------------------------------------
;; Comments
;; -----------------------------------------------------------------------------
(defun manual-get-couple-comment (couple)
  ""
  (if (and manual-alignment
	   (<= 0 couple))
      (nth 4 (nth couple (manual-get-couple-list)))))


(defun manual-set-comment (comment)
  ""
  (interactive (list (read-string "Comment: " (or (manual-get-couple-comment manual-couple)
						  (error "No current alignment")))))
  (manual-set-modified-flag t)
  (setcar (nthcdr 4 (nth manual-couple (manual-get-couple-list))) comment)
  (manual-update-modelines))


(defun manual-show-comment ()
  ""
  (interactive)
  (if (not manual-alignment)
      (error "No alignment"))
  (message "Comment: %s"
	   (or (manual-get-couple-comment manual-couple)
	       (error "No current alignment"))))


;; =============================================================================
;; Navigation
;; =============================================================================

(defun manual-goto-first-couple ()
  ""
  (interactive)
  (if (not manual-alignment)
      (error "No alignment"))
  (setq manual-couple 0)
  (manual-update-display))


(defun manual-goto-last-couple ()
  ""
  (interactive)
  (if (not manual-alignment)
      (error "No alignment"))
  (setq manual-couple (1- (length (manual-get-couple-list))))
  (manual-update-display))


(defun manual-goto-next-couple (n)
  ""
  (interactive "p")
  (if (not manual-alignment)
      (error "No alignment"))
  (if (not n) 
      (setq n 1))
  (setq manual-couple (mod (+ manual-couple n) (length (manual-get-couple-list))))
  (manual-update-display))


(defun manual-goto-previous-couple (n)
  ""
  (interactive "p")
  (if (not manual-alignment)
      (error "No alignment"))
  (manual-goto-next-couple (- n)))


(defun manual-goto-nth-couple (n)
  ""
  (interactive "NCouple no.: ")
  (if (not manual-alignment)
      (error "No alignment"))
  (if (or (< n 1)
	  (> n (length (manual-get-couple-list))))
      (error "manual-goto-nth-couple: %d out of range [1, %d]"
	     n (length (manual-get-couple-list))))
  (setq manual-couple (1- n))
  (manual-update-display))


(defun manual-goto-next-comment (n)
  "Find and go to the next commented couple (previous with negative arg)"
  (interactive "p")

  (if (not manual-alignment)
      (error "No alignment"))
  (let ((increment (if (< n 0) -1 1))
	(current manual-couple)
	(max (length (manual-get-couple-list))))
    (while (and (/= manual-couple (setq current (mod (+ current increment) max)))
		(string= "" (manual-get-couple-comment current))))
    (if (string= "" (manual-get-couple-comment current))
	(error "No comments"))
    (manual-goto-nth-couple (1+ current))
    (manual-show-comment)))


(defun manual-goto-previous-comment (n)
  "Find and go to the previous commented couple (next with negative arg)"
  (interactive "p")

  (manual-goto-next-comment (- n)))    
    


(defvar manual-previous-search-string nil)

(defun manual-search-for-comment (reverse search-string)
  "Search for and go to the next couple (previous with prefix argument) 
whose comment matches regexp SEARCH-STRING (repeat previous search if
SEARCH-STRING is an empty string)"
  (interactive "P\nsComment search: ")

  (if (not manual-alignment)
      (error "No alignment"))

  (if (and (string= search-string "") manual-previous-search-string)
      (setq search-string manual-previous-search-string))

  (setq manual-previous-search-string search-string)

  (let ((increment (if reverse -1 1))
	(current manual-couple)
	(max (length (manual-get-couple-list)))
	(match nil))
    (while (and (/= manual-couple (setq current (mod (+ current increment) max)))
		(not (setq match (string-match search-string (manual-get-couple-comment current))))))
    (if (not match)
	(error "No matches")
      (if (< current manual-couple)
	  (message "Wrap-around...")))
    (manual-goto-nth-couple (1+ current))
    (manual-show-comment)))


(defun manual-search-for-couple-text (reverse search-string)
  "Search for and go to the next couple (previous with prefix argument) 
whose text matches regexp SEARCH-STRING (repeat previous search if
SEARCH-STRING is an empty string)"
  (interactive "P\nsCouple search: ")

  (if (not manual-alignment)
      (error "No alignment"))

  (if (and (string= search-string "") manual-previous-search-string)
      (setq search-string manual-previous-search-string))

  (setq manual-previous-search-string search-string)

  (let ((increment (if reverse -1 1))
	(current manual-couple)
	(max (length (manual-get-couple-list)))
	(match nil))
    
    (setq match (if reverse (re-search-backward search-string)
		  (re-search-forward search-string)))
    (manual-find-couple)
    (goto-char match)))


(defun manual-search-for-diff (reverse)
  "Search for next '<diff-begin' mark in couple comments"
  (interactive "P")
  (let (manual-previous-search-string)
    (manual-search-for-comment reverse "<diff-begin [0-9]+>")))


(defun manual-search-for-diff-end ()
  "Search for 'diff-end' mark corresponding to this 'diff-begin' in couple comments"
  (interactive)
  (if (not manual-alignment)
      (error "No alignment"))

  (let (manual-previous-search-string diff-no)
    (if (not (string-match "<diff-begin \\([0-9]+\\)>" (manual-get-couple-comment manual-couple)))
	(error "Not on a diff-begin!"))
    (setq diff-no (match-number 1 (manual-get-couple-comment manual-couple)))
    (message "Searching for end of difference %d..." diff-no)
    (if (string-match (format "<diff-end %d>" diff-no) (manual-get-couple-comment manual-couple))
	(message "Difference %d begins and ends here" diff-no)
      (manual-search-for-comment nil (format "<diff-end %d>" diff-no))
      (message "End of difference %d" diff-no))))


(defun manual-find-couple ()
  "Find the couple within which point lies"

  (interactive)

  (if (not manual-alignment)
      (error "No alignment"))
  (let ((buf-index (cond ((equal (current-buffer) (manual-get-buffer 1))
			  1)
			 ((equal (current-buffer) (manual-get-buffer 2))
			  2)
			 (t
			  (error "buffer %s is not an aligned buffer"
				 (buffer-name (current-buffer))))))
	(min-couple 0)
	(max-couple (1- (length (manual-get-couple-list))))
	couple
	(position (point))
	(buffer (current-buffer)))

    (setq couple (/ max-couple 2))
    (while (and (or (< position (manual-get-couple-start couple buf-index))
		    (>= position (manual-get-couple-end couple buf-index)))
		(> (- max-couple min-couple) 5))
      (cond ((< position (manual-get-couple-start couple buf-index))
	     (setq max-couple (1- couple)))
	    ((>= position (manual-get-couple-end couple buf-index))
	     (setq min-couple (1+ couple))))
      (setq couple (+ (/ (- max-couple min-couple) 2) min-couple)))
      
    (setq couple min-couple)
    (while (and (< couple max-couple)
		(>= position (manual-get-couple-end couple buf-index)))
      (setq couple (1+ couple)))

    (if (/= couple manual-couple)
	(manual-goto-nth-couple (1+ couple)))))	;Careful! goto-nth substracts 1...


(defun manual-mouse-goto-couple (event)
  ""
  (interactive "e")
  (let* ((window (nth 0 (nth 1 event)))
	 (position (nth 1 (nth 1 event)))
	 (buffer (window-buffer window))
	 (half (if (equal buffer (manual-get-buffer 1)) 1 2)))
    (set-window-point window position) 
    (select-window window)
    
    (manual-find-couple)

    (manual-move-to-closest-boundary buffer position
				     (manual-get-couple-start manual-couple half)
				     (manual-get-couple-end manual-couple half))))


;; =============================================================================
;; Alignment modification
;; =============================================================================

(defun manual-split-couple ()
  "Split the current couple of alignment at current point positions"
  (interactive)
  (if (not manual-alignment)
      (error "No alignment"))
  (let* ((tail (nthcdr manual-couple (manual-get-couple-list)))
	 couple pos1 pos2)

    (save-excursion
      (set-buffer (manual-get-buffer 1))
      (setq pos1 (point))
      (if (or (< pos1 (manual-get-couple-start manual-couple 1))
	      (> pos1 (manual-get-couple-end manual-couple 1)))
	  (error "Position %d not in current couple" pos1))

      (set-buffer (manual-get-buffer 2))
      (setq pos2 (point))
      (if (or (< pos2 (manual-get-couple-start manual-couple 2))
	      (> pos2 (manual-get-couple-end manual-couple 2)))
	  (error "Position %d not in current couple" pos2))

      (if (or (and (= pos1 (manual-get-couple-start manual-couple 1))
		   (= pos2 (manual-get-couple-start manual-couple 2)))
	      (and (= pos1 (manual-get-couple-end manual-couple 1))
		   (= pos2 (manual-get-couple-end manual-couple 2))))
	  (error "Split would generate an empty couple")))

    (manual-set-modified-flag t)
    (setq couple (car tail))
    (setcdr tail (cons (list (1- pos1) (nth 1 couple)
			     (1- pos2) (nth 3 couple)
			     "")
		       (cdr tail)))	; Add new couple

    (setq couple (car tail))		; Modify ends of current couple
    (setcar (nthcdr 1 couple) (1- pos1))
    (setcar (nthcdr 3 couple) (1- pos2))

    (setq manual-couple (1+ manual-couple))) ; Make new couple current
  
  (manual-update-display)
  (manual-move-to-next-boundary (manual-get-buffer 1)
				(manual-get-couple-start manual-couple 1)
				(min (manual-get-couple-end manual-couple 1)
				     (+ 500 (manual-get-couple-start manual-couple 1))))
  (manual-move-to-next-boundary (manual-get-buffer 2)
				(manual-get-couple-start manual-couple 2)
				(min (manual-get-couple-end manual-couple 2)
				     (+ 500 (manual-get-couple-start manual-couple 2)))))
    
(defun manual-merge-couples (n)
  "Merge the current couple with the N following ones (preceding if N is negative)"
  (interactive "p")

  (if (not manual-alignment)
      (error "No alignment"))
  (if (= n 0)
      (error "Invalid prefix argument for merge: %d" n))
  (if (< n 0)
      (if (> (- n) manual-couple)
	  (error "Not enough couples before current for merge!")
	(setq manual-couple (+ manual-couple n))
	(setq n (- n))))
  (let ((tail (nthcdr manual-couple (manual-get-couple-list)))
	this next)
    (if (not (nthcdr n tail))
	(error "Not enough couples after current for merge!"))
    (manual-set-modified-flag t)
    (setq this (car tail))
    (setq next (car (nthcdr n tail)))
    (setcar (nthcdr 1 this) (nth 1 next))		; Modify ends of current couple
    (setcar (nthcdr 3 this) (nth 3 next))
    (setcdr tail (nthcdr (1+ n) tail)))
  (manual-update-display))


(defun manual-merge-couples-backward (n)
  "Merge the current couple with the N preceding ones (following if N is negative)"
  (interactive "p")

  (manual-merge-couples (- n)))


(defun manual-set-modified-flag (flag)
  ""
  (if (not manual-alignment)
      (error "No alignment"))
  (setcar (nthcdr 4 manual-alignment) flag))



;; =============================================================================
;; Initial setup, I/O functions
;; =============================================================================

(defun manual-initialize (alignment)
  "Make ALIGNMENT the current alignment, and perform various
initializations."
  (setq manual-alignment alignment)
  (setq manual-couple 0)
  (manual-update-display))


(defun manual-load (text1-file text2-file alignment-file)
  "Load a pair of text files and an alignment"
  (interactive (progn
		  (if (and manual-alignment
		           (not (manual-quit)))
		      (error "You must quit before loading another alignment"))
		  (list (read-file-name "French text file: " nil nil t nil)
			(read-file-name "English text file: " nil nil t nil)
			(read-file-name "Alignment file: " nil nil t nil))))

  (let (buf1 buf2 couples new-al)
    (setq buf1 (find-file text1-file))
    (save-excursion 
      (set-buffer buf1)
      (manual-mode)
      (setq buffer-read-only t)
      (auto-save-mode nil))

    (setq buf2 (find-file-other-window text2-file))
    (save-excursion 
      (set-buffer buf2)
      (manual-mode)
      (setq buffer-read-only t)
      (auto-save-mode nil))

    (setq couples (manual-read-couples alignment-file))

    (setq new-al (list buf1 buf2 couples alignment-file nil nil))
    (manual-check-alignment new-al)

    (manual-initialize new-al))
  (manual-update-display)
  (message "Done."))



(defun manual-new (text1-file text2-file)
  "Load a pair of text files and create a new alignment"
  (interactive (progn
		 (if (and manual-alignment
			  (not (manual-quit)))
		     (error "You must quit before creating a new alignment"))
		 (list (read-file-name "French text file: " nil nil t nil)
		       (read-file-name "English text file: " nil nil t nil))))

  (let (buf1 buf2 couples new-al)
    (setq buf1 (find-file text1-file))
    (save-excursion 
      (set-buffer buf1)
      (manual-mode)
      (setq buffer-read-only t)
      (auto-save-mode nil))

    (setq buf2 (find-file-other-window text2-file))
    (save-excursion 
      (set-buffer buf2)
      (manual-mode)
      (setq buffer-read-only t)
      (auto-save-mode nil))

    (setq couples (list (list 0 (size-of-buffer buf1) 0 (size-of-buffer buf2) "")))

    (setq new-al (list buf1 buf2 couples nil t nil))

    (manual-initialize new-al))
  (manual-update-display)
  (message "Done."))



;;;;(defconst manual-entry-RE
;;;;  "<align x0=\\([0-9]+\\) x1=\\([0-9]+\\) y0=\\([0-9]+\\) y1=\\([0-9]+\\) comment=\"\\([^\"]*\\)\">")
(defconst manual-entry-RE
  "^[ \t]*\\([0-9]+\\)[ \t]+\\([0-9]+\\)\\([ \t]+\\(.*\\)\\)?$")


(defun manual-read-couples (file)
  "Visit FILE, and try to read its content as a list of couples in 'manual'
format.  Upon success, return the list as a Lisp object, nil otherwise"
  (let (couples
	(count 0)
	(x0 0)
	(y0 0))
    (save-excursion
      (set-buffer (get-buffer-create "*manual-tmp-buffer*"))
      (insert-file-contents file nil nil nil 'replace)

      (goto-char (point-min))
      (while (re-search-forward manual-entry-RE nil t)
	(setq count (1+ count))
	(if (= 0 (mod count 50)) (message "Parsing alignment entries: %d" count))
	(setq couples (append couples (list (list x0 (match-number 1)
						  y0 (match-number 2)
						  (match-substring 4)))))
	(setq x0 (match-number 1))
	(setq y0 (match-number 2)))
      (if (not couples)
	  (error "No couples in %s" file))
      (set-buffer-modified-p nil)
      (kill-buffer "*manual-tmp-buffer*"))
    couples))



(defun manual-set-save-to-file (filename)
  ""
  (interactive "FWrite alignment to file: ")
  (if (not (file-writable-p filename))
      (error "Can't create or write to file %s" filename))
  (setcar (nthcdr 3 manual-alignment) filename)
  (manual-set-modified-flag t)
  (manual-update-modelines))
  
    

(defun manual-write (filename)
  "Write the alignment to a new file; associate this file with aligment"
  (interactive "FWrite alignment to file: ")
  (manual-set-save-to-file filename)  
  (manual-save 'no-confirm))


(defun manual-save (&optional no-confirm)
  "Save the alignment onto the associated file; query the user for a file
name if this is nil"
  (interactive "P")
  
  (if (not manual-alignment)
      (error "No alignment to save"))

  (if (or no-confirm
	  (y-or-n-p "Save alignment? "))
      (let (couples couple)
	
	(if (not (manual-get-save-to-file))
	    (call-interactively 'manual-set-save-to-file))

	(if (manual-get-modified-flag)
	    (let (couples couple)
	      (save-excursion
		(set-buffer (get-buffer-create "*manual-tmp-buffer*"))
		(erase-buffer)

		(setq couples (manual-get-couple-list))
		(while couples
;;;;	    (insert (format "<align x0=%d x1=%d y0=%d y1=%d comment=\"%s\">\n"
;;;;			    (nth 0 (car couples))
;;;;			    (nth 1 (car couples))
;;;;			    (nth 2 (car couples))
;;;;			    (nth 3 (car couples))
;;;;			    (nth 4 (car couples))))
		  (insert (format "%d\t%d\t%s\n"
				  (nth 1 (car couples))
				  (nth 3 (car couples))
				  (nth 4 (car couples))))
			    
		  (setq couples (cdr couples)))

		(let ((filename (expand-file-name (manual-get-save-to-file))))
		  (if (and filename
			   (file-exists-p filename)
			   (not (manual-get-backed-up-flag)))
		      (cond ((/= 0 (call-process "cp" nil nil nil 
						 filename
						 (make-backup-file-name filename)))
			     (beep) 
			     (message "Can't backup %s" filename))
			    (t 
			     (manual-set-backed-up-flag t))))
		  (write-region (point-min) (point-max) filename nil 'no-message))

		(set-buffer-modified-p nil)
		(kill-buffer "*manual-tmp-buffer*"))
	      (message "Wrote alignment to file %s." (manual-get-save-to-file)))
	  (message "No changes need to be saved."))
	(manual-set-modified-flag nil)
	(manual-set-backed-up-flag t))))



(defun manual-check-alignment (alignment)
  "Verify some basic properties of an alignment."

  t)


(defun manual-quit (&optional no-confirm)
  "Quit Manual"
  (interactive "P")
  (cond ((and manual-alignment
	      (or no-confirm 
		  (y-or-n-p "Quit Manual? ")))
	 (if (manual-get-modified-flag)
	     (manual-save no-confirm))
	 (save-excursion
	   (manual-clear-display)
	   (set-buffer (manual-get-buffer 1))
	   (set-buffer-modified-p nil)
	   (normal-mode)
	   (set-buffer (manual-get-buffer 2))
	   (set-buffer-modified-p nil)
	   (normal-mode)
	   (setq manual-alignment nil)
	   (setq manual-couple 0)
	   (manual 'disactivate)))
	(t nil)))
	

(defun manual-set-backed-up-flag (flag)
  ""
  (if (not manual-alignment)
      (error "No alignment"))
  (setcar (nthcdr 5 manual-alignment) flag))



;; =============================================================================
;; Display functions
;; =============================================================================


;; -----------------------------------------------------------------------------
;; Text-properties, colors, etc.
;; -----------------------------------------------------------------------------

(defvar manual-highlight-faces nil)
(defvar manual-highlight-colors nil)

(cond ((and window-system (x-display-color-p))
       (setq manual-highlight-faces
	 '(manual-face-2
	   manual-face-1
	   manual-face+0
	   manual-face+1
	   manual-face+2))

       (setq manual-highlight-colors
	 '("darkseagreen3"
           "darkseagreen1"
           "lemonchiffon"
           "lightsteelblue1"
           "lightsteelblue3"))

       (set-face-background (make-face 'manual-sentence-end-face) "burlywood3")

       (let ((faces manual-highlight-faces)
	     (colors manual-highlight-colors))
	 (while faces
	   (make-face (car faces))
	   (set-face-background (car faces) (car colors))
	   (setq faces (cdr faces))
	   (setq colors (cdr colors)))))
      (t 
       (error "Got to do something about dumb terminals...")))




;; -----------------------------------------------------------------------------
;; Update the highlighted region
;; -----------------------------------------------------------------------------
(defun manual-clear-display ()
  ""
  (if manual-alignment
      (save-excursion
	(let ((inhibit-read-only t))
	  (set-buffer (manual-get-buffer 1))
	  (remove-text-properties (point-min) (point-max) '(face nil))

	  (set-buffer (manual-get-buffer 2))
	  (remove-text-properties (point-min) (point-max) '(face nil))))))


(defun manual-update-display ()
  "Update the display so that the current Manual alignment is visible"
  (if manual-alignment
      (let ((inhibit-read-only t) 
	    couple faces)
	(save-excursion
	  (manual-clear-display)
	  (setq faces manual-highlight-faces)
	  (setq couple (- manual-couple (/ (length faces) 2)))
	  (while faces
	    (if (manual-get-couple-start couple 1)
		(set-text-properties (manual-get-couple-start couple 1)
				     (manual-get-couple-end couple 1)
				     (list 'face (car faces))
				     (manual-get-buffer 1)))
	    (setq couple (1+ couple))
	    (setq faces (cdr faces)))
	  (manual-highlight-sentence-ends (manual-get-buffer 1)
					  (manual-get-couple-start manual-couple 1)
					  (min (manual-get-couple-end manual-couple 1)
					       (+ 1000 (manual-get-couple-start manual-couple 1)))
					  manual-number-sentence-ends)
				       

	  (setq faces manual-highlight-faces)
	  (setq couple (- manual-couple (/ (length faces) 2)))
	  (while faces
	    (if (manual-get-couple-start couple 2)
		(set-text-properties (manual-get-couple-start couple 2)
				     (manual-get-couple-end couple 2)
				     (list 'face (car faces))
				     (manual-get-buffer 2)))
	    (setq couple (1+ couple))
	    (setq faces (cdr faces)))
	  (manual-highlight-sentence-ends (manual-get-buffer 2)
					  (manual-get-couple-start manual-couple 2)
					  (min (manual-get-couple-end manual-couple 2)
					       (+ 1000 (manual-get-couple-start manual-couple 2)))
					  manual-number-sentence-ends))

	(manual-update-modelines)

	(manual-show-buffers))))

(defun manual-update-modelines ()
  ""
  (let ((buf-id (list (format "Manual: %%9b %9s: %s%d of %d"
			      (if (manual-get-save-to-file)
				  (file-name-nondirectory (manual-get-save-to-file)) 
				"(new alignment)")
			      (if (< 0 (length (manual-get-couple-comment manual-couple))) "*" " ")
			      (1+ manual-couple)
			      (length (manual-get-couple-list))))))
    (set-buffer (manual-get-buffer 1))
    (setq mode-line-buffer-identification buf-id)
    (set-buffer (manual-get-buffer 2))
    (setq mode-line-buffer-identification buf-id)
    (force-mode-line-update))) 

(defun manual-show-buffers ()
  "Set up the Emacs display to show the Manual buffers"
  (let ((text1-window (get-buffer-window (manual-get-buffer 1)))
	(text2-window (get-buffer-window (manual-get-buffer 2)))
	(start1 (manual-get-couple-start manual-couple 1))
	(end1 (manual-get-couple-end manual-couple 1))
	(start2 (manual-get-couple-start manual-couple 2))
	(end2 (manual-get-couple-start manual-couple 2)))
    (cond ((not (window-live-p text1-window))
	   (setq text1-window nil)))
    (cond ((not (window-live-p text2-window))
	   (setq text2-window nil)))
    (cond ((not (and text1-window text2-window))
	   (cond ((one-window-p t "anything else")
		  (split-window)))
	   (cond ((not text1-window)
		  (setq text1-window (cond (text2-window
					   (next-window text2-window "no minibuf"))
					  (t
					   (selected-window))))
		  (set-window-buffer text1-window (manual-get-buffer 1))))
	   (cond ((not text2-window)
		  (setq text2-window (next-window text1-window "no minibuf"))
		  (set-window-buffer text2-window (manual-get-buffer 2))))))
    (cond (window-system
	   (if (or (< (window-point text1-window) start1)
		   (>= (window-point text1-window) end1))
	       (progn
		 (set-window-point text1-window start1)
		 (display-region text1-window start1 end1)))
	   (if (or (< (window-point text2-window) start2)
		   (>= (window-point text2-window) end2))
	       (progn
		 (set-window-point text2-window start2)
		 (display-region text2-window start2 end2))))
	  (t 
	   (set-window-point text1-window pos1)
	   (set-window-start text1-window pos1 nil)
	   (set-window-point text2-window pos2)
	   (set-window-start text2-window pos2 nil)))))



;; -----------------------------------------------------------------------------
;; Potential segmentation points
;; -----------------------------------------------------------------------------

(defvar manual-number-sentence-ends 10
  "*In Manual-mode, the maximum number of potential sentence ends to be 
highlighted in the current alignment couple.")

(defconst manual-sentence-end-RE "\\([.?!;]\\s-\\|\f\\|\n\\s-*\n\\)")

(defun manual-highlight-sentence-ends (buffer start end max)
  "Find and highlight at most MAX potential sentence ends in BUFFER,  between
positions START and END"
  (save-excursion
    (let ((count 0)
	  match-position)
      (set-buffer buffer)
      (remove-text-properties (point-min) (point-max) '(sentence-boundary t) buffer)
      (goto-char start)
      (if (looking-at "\\s-+")
	  (goto-char (match-end 0)))
      (while (and (< count max)
		  (setq match-position (manual-find-sentence-end (max (point) end))))
	(set-text-properties match-position
			     (min (1+ match-position) (point-max))
			     (list 'face 'manual-sentence-end-face 
				   'sentence-boundary t)
			     buffer)
	(setq count (1+ count))))))
		  

(defun manual-find-sentence-end (limit)
  "Find the next potential sentence end point"
  
  (save-match-data
    (if (re-search-forward manual-sentence-end-RE (max (point) end) 'move-to-limit-on-failure)
	(let ((position (match-beginning 0)))
	  (save-excursion
	    (goto-char position)
	    (if (re-search-forward "\\s-" (match-end 0) t)
		(setq position (match-beginning 0))))
	  position)
      nil)))
	

(defun manual-move-to-closest-boundary (buffer pos min max)
  ""
  (set-buffer buffer)
  (if (get-text-property pos 'sentence-boundary buffer)
      (goto-char pos)
    (if (get-text-property (1- pos) 'sentence-boundary buffer)
	(goto-char (1- pos))
      (let (previous next)
    
	(setq previous (previous-single-property-change pos 'sentence-boundary
							buffer min))
	(setq next (next-single-property-change pos 'sentence-boundary
						buffer max))
	(goto-char (if (< (- pos previous) (- next pos))
		       (max min (1- previous))
		     next))))))

(defun manual-move-to-next-boundary (buffer pos max)
  ""
  (interactive (let ((half (cond ((equal (current-buffer) (manual-get-buffer 1)) 1)
				 ((equal (current-buffer) (manual-get-buffer 2)) 2)
				 (t (error "Not an aligned buffer")))))
		 (list (set-buffer (current-buffer))
		       (point)
		       (manual-get-couple-end manual-couple half))))

  (set-buffer buffer)
  (if (get-text-property pos 'sentence-boundary buffer)
      (setq pos (next-single-property-change pos 'sentence-boundary
					     buffer max)))
  (setq pos (next-single-property-change pos 'sentence-boundary
					 buffer max))
  (if (get-text-property pos 'sentence-boundary buffer)
      (progn 
	(goto-char pos)
	(if (get-buffer-window buffer 'visible) 
	    (set-window-point (get-buffer-window buffer 'visible) pos)))))



;; =============================================================================
;; Manual major mode
;; =============================================================================


(defvar manual-mode-hook nil)
(defvar manual-mode-map nil)
(if manual-mode-map
    nil
  (setq manual-mode-map (make-keymap))
  (suppress-keymap manual-mode-map)
  (define-key manual-mode-map "g" 'manual-goto-nth-couple)
  (define-key manual-mode-map [down-mouse-1] 'manual-mouse-goto-couple)
  (define-key manual-mode-map [drag-mouse-1] 'manual-mouse-goto-couple)
  (define-key manual-mode-map [mouse-1] 'manual-mouse-goto-couple)
  (define-key manual-mode-map "n" 'manual-goto-next-couple)
  (define-key manual-mode-map "p" 'manual-goto-previous-couple)
  (define-key manual-mode-map "f" 'manual-goto-first-couple)
  (define-key manual-mode-map "l" 'manual-goto-last-couple)
  (define-key manual-mode-map "c" 'manual-show-comment)
  (define-key manual-mode-map "C" 'manual-set-comment)
  (define-key manual-mode-map ";" 'manual-goto-next-comment)
  (define-key manual-mode-map ":" 'manual-goto-previous-comment)
  (define-key manual-mode-map "S" 'manual-search-for-comment)
  (define-key manual-mode-map "t" 'manual-search-for-couple-text)
  (define-key manual-mode-map "d" 'manual-search-for-diff)
  (define-key manual-mode-map "D" 'manual-search-for-diff-end)
  
  (define-key manual-mode-map " " 'manual-split-couple)
  (define-key manual-mode-map "m" 'manual-merge-couples)
  (define-key manual-mode-map "u" 'manual-merge-couples-backward)
  (define-key manual-mode-map "s" 'manual-save)
  (define-key manual-mode-map "w" 'manual-write)
  (define-key manual-mode-map "r" 'manual-load)
  (define-key manual-mode-map "N" 'manual-new)
  (define-key manual-mode-map "q" 'manual-quit)

  (define-key manual-mode-map "\t" 'manual-move-to-next-boundary)

  (define-key manual-mode-map "o" 'other-window))

(defun manual-mode ()
  "Manual mode

This major-mode is used to view, create or modify bitext alignments.  The
command 'M-x manual' adds a 'Manual' menu to the Emacs menubar.  Most
functions are available through this menu.  Here are some of the more
interesting key-bindings:

Key     Command				Description
---     -------				-----------
q	manual-quit			Quit Manual-mode
. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . 
f	manual-goto-first-couple	Go to first couple of alignment
n	manual-goto-next-couple		Go to next couple of alignment
p	manual-goto-previous-couple	Go to previous couple of
					alignment 
l	manual-goto-last-couple		Go to last couple of alignment
g	manual-goto-nth-couple		Go to some specific couple
TAB	manual-move-to-next-boundary	Move point to next potential
					sentence boundary in couple
o	other-window			select other window (you'll appreciate
					having a quick binding for this one...)
. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . 
c	manual-show-comment		Show couple's associated comment (in the
					minibuffer)
C	manual-set-comment		Edit couple's associated comment
;	manual-goto-next-comment	Go to next couple that has an associated
					comment
:	manual-goto-previous-comment	Go to previous such couple
S	manual-search-for-comment	Search for next comment that matches
					a given regular-expression
t	manual-search-for-couple-text	Do a regular expression search in
					the current aligned buffer, highlighting
					matching couple in both buffers
. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . 
SPC	manual-split-couple		Make two couples from the current
					couple, splitting at current point 
					positions
m	manual-merge-couples		Merge the current couple with next
u	manual-merge-couples-backward	Merge the current couple with previous
					(hint: after a split, this is like an
					\"undo\")
. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . 
s	manual-save			Save the current alignment
w	manual-write			Save alignment to a new file
r	manual-load			Load an alignment
N	manual-new			Create a new alignment

Also, clicking the left mouse-button at some position in one of the aligned
buffers will \"go to\" the corresponding couple, and move point to the 
closest potential sentence-boundary (i.e. call the 'manual-mouse-find-couple' 
command).  To move point to a position other than a sentence boundary,
double-click on that position, or use the arrow keys.

"
  (interactive)
  (kill-all-local-variables)
  (setq major-mode 'manual-mode)
  (setq mode-name "Manual")
  (put 'manual-mode 'mode-class 'special)
  (use-local-map manual-mode-map)
  (run-hooks 'manual-mode-hook))


;; -----------------------------------------------------------------------------
;; Menus
;; -----------------------------------------------------------------------------

(defvar manual-menu-map nil)

;(setq manual-menu-map nil)		;;;;; DEBUGGING

;; Menu bindings
(defun manual-menu-separator ()
  ""
  (interactive))

(defun manual (&optional disactivate)
  "Activate or disactivate the Manual menu and functions"
  (interactive "P")
  (if (and (not disactivate)
	   (not manual-menu-map))
      (progn
	(setq manual-menu-map (make-sparse-keymap "Manual"))

	(define-key manual-menu-map [quit]
	  '("Quit Manual" . manual-quit)) 
	(define-key manual-menu-map [sep5]
	  '("--" . manual-menu-separator))
	(define-key manual-menu-map [merge-couples-backward]
	  '("Merge with Previous" . manual-merge-couples-backward)) 
	(define-key manual-menu-map [merge-couples]
	  '("Merge with Next" . manual-merge-couples)) 
	(define-key manual-menu-map [split-couple]
	  '("Split Couple" . manual-split-couple)) 
	(define-key manual-menu-map [sep4]
	  '("--" . manual-menu-separator))
	(define-key manual-menu-map [search-for-comment]
	  '("Search for Comment..." . manual-search-for-comment)) 
	(define-key manual-menu-map [previous-comment]
	  '("Goto Previous Comment" . manual-goto-previous-comment)) 
	(define-key manual-menu-map [next-comment]
	  '("Goto Next Comment" . manual-goto-next-comment)) 
	(define-key manual-menu-map [set-comment]
	  '("Edit Comment" . manual-set-comment)) 
	(define-key manual-menu-map [show-comment]
	  '("Show Comment" . manual-show-comment)) 
	(define-key manual-menu-map [sep3]
	  '("--" . manual-menu-separator))
	(define-key manual-menu-map [search-for-couple]
	  '("Search..." . manual-search-for-couple-text)) 
	(define-key manual-menu-map [nth-couple]
	  '("Goto Couple no..." . manual-goto-nth-couple)) 
	(define-key manual-menu-map [last-couple]
	  '("Last Couple" . manual-goto-last-couple)) 
	(define-key manual-menu-map [previous-couple]
	  '("Previous Couple" . manual-goto-previous-couple)) 
	(define-key manual-menu-map [next-couple]
	  '("Next Couple" . manual-goto-next-couple)) 
	(define-key manual-menu-map [first-couple]
	  '("First Couple" . manual-goto-first-couple)) 
	(define-key manual-menu-map [sep2]
	  '("--" . manual-menu-separator))
	(define-key manual-menu-map [write-align]
	  '("Save Alignment as..." . manual-write)) 
	(define-key manual-menu-map [save-align]
	  '("Save Alignment" . manual-save)) 
	(define-key manual-menu-map [load-align]
	  '("Load Alignment..." . manual-load)) 
	(define-key manual-menu-map [new]
	  '("New Alignment..." . manual-new)) 

	(setq menu-bar-final-items (cons 'manual menu-bar-final-items))))

  (define-key global-map [menu-bar manual] (cons "Manual" manual-menu-map)))

(add-hook 'kill-emacs-hook 'manual-quit)

;; =============================================================================
;; Utilities
;; =============================================================================
;; -----------------------------------------------------------------------------
;; read-clicked-position
;; -----------------------------------------------------------------------------

(defun read-clicked-position (buffer &optional prompt)
  "Prompt the user to click the mouse in buffer, and return the clicked position"
  
  (setq buffer (buffer-name (get-buffer buffer))) ; So we are sure we have a buffer-name

  (if (not (get-buffer-window buffer 'visible))
      (error "read-clicked-position: buffer %s is not visible" buffer))

  (if (not prompt)
      (setq prompt (format "Please click a position in buffer %s" buffer)))
  (let ((count 0)
	event window position)

    (while (not (and (message "%s" prompt)
		     (setq event (read-event))
		     (or (eq 'mouse-1 (event-basic-type event))
			 (eq 'mouse-2 (event-basic-type event))
			 (eq 'mouse-3 (event-basic-type event)))
		     (member 'click (event-modifiers event))
		     (string= buffer (buffer-name (window-buffer (car (nth 1 event)))))))
      (setq count (1+ count))
      (if (= (% count 4) 0) (beep t)))

    (setq window (nth 0 (nth 1 event)))
    (setq position (nth 1 (nth 1 event)))
    (set-window-point window position)
    (select-window window)
    position))



;; -----------------------------------------------------------------------------
;; Display stuff
;; -----------------------------------------------------------------------------
(defun display-region (window start end)
  "Center window display around region delimited by start and end"
  (cond ((not (and (my-pos-is-visible window start)
		   (my-pos-is-visible window end)))
	 (let ((current-selected-window (selected-window)))
	   (select-window window)
	   (set-buffer (window-buffer window))
	   (setq start (max start (point-min)))
	   (setq end (min end (point-max)))
	   (goto-char start)
	   (recenter (max 1 (/ (- (window-height window)
				  (count-lines start end))
			       2)))
	   (select-window current-selected-window)))))


(defun my-pos-is-visible (window pos)
  "Return true if pos is visible within window"
  (and (<= (window-start window) pos)
       (<  pos (window-end window))))






;; -----------------------------------------------------------------------------
;; Match-data stuff
;; -----------------------------------------------------------------------------
(defun match-substring (count)
  "Return the COUNTth item of current match-data as a string"
  
  (let ((start (match-beginning count))
	(end (match-end count)))
    (if (and start end)
	(buffer-substring start end)
      "")))

(defun match-number (count &optional string)
  "Return the COUNTth item of current match-data as a number"
  
  (let ((start (match-beginning count))
	(end (match-end count)))
    (if (not (and start end))
	(error "match-integer: COUNT=%d too large" count))
    (string-to-number (if string (substring string start end)
			(buffer-substring start end)))))


;; -----------------------------------------------------------------------------
;; Various
;; -----------------------------------------------------------------------------

(defun size-of-buffer (buffer)
  "Return the number of characters in a given buffer"
  (save-excursion
    (set-buffer buffer)
    (buffer-size)))




;; =============================================================================
;; Init
;; =============================================================================
(manual)
